
# [WIP] Ez DMA Helper Library

Many microcontrollers / System-on-Chip systems have DMA-enabled peripherals
which allow offloading processing from the main CPU(s).  However, use of
such peripherals (especially with real-time requirements) can be difficult.

This library is aimed at developers of libraries that simplify the use of
these peripherals, by simplifying the task of managing the buffers used
when interacting with these devices.

## Example difficulties

Compared to having a CPU-based task processesing data buffers, many
DMA-enabled peripherals act ... differently.  As examples, they may:

* double-buffer the address to DMA to in the peripheral's internal registers.
* require that the _NEXT_ buffer pointer is written, even while the current DMA transaction is taking place.
* therefore, they may not even have a register that represent, when 
  an event indicating that DMA is finished, which buffer pointer it
  was that the event refers to.

## Features

Selected features of the EzDmaHelper class include:

* Template parameters used to allow compile-time optimizations critical to IoT-class CPUs
  * Template parameter to support arbitrary per-buffer size
  * Template parameter to support arbitrary count of buffers (<= 255)
  * Template parameter to support arbitrary alignment requirement
* Provides event callbacks at arbitrary priority (via FreeRTOS task)
* 100% static allocation ... no heap fragmentation over long use

## Background and other notes

When designing a producer/consumer abstraction that only deals with CPU-based
tasks, the pointer currently being worked on by a given CPU-based task is
always available.  Thus, when the task is finished writing to a buffer,
it's easy to send that buffer pointer over to the other task.

Things get a little trickier when dealing with DMA-based peripheral hardware.
This is especially true when there are real-time requirements, as the
peripheral needs the next buffer's pointer as soon as it completes the prior
transfer.  Therefore, it cannot wait for an interrupt handler to fire, and
wait for the interrupt handler to write a new buffer pointer.  This is
especially true when (for example) additional latency between CPU and peripheral
exists.

### Example hardware - nRF52 peripherals

The peripheral hardware's does two things:

1. double-buffers registers that store the address(es) used for DMA
2. generates an event (interrupt) when the register is copied to the peripheral's internal location (to indicate it's safe to update)
3. generates an event (interrupt) when a DMA transaction completes

This means that, when the interrupt handler handles the first event,
the interrupt handler can signal the software that it is safe to updates
the register with the address of the buffer pointer for the **_NEXT_** DMA
operation.  In this way, when the peripheral reaches the end of the first
buffer, it can immediately begin transferring data to the NEXT buffer,
without any delay waiting on the CPU to handle an interrupt, and without
any delays as registers as read/written from the CPU.

### Example Sequence

So, what's this sequence look like?  Let's consider three buffers being used
by a DMA peripheral, with data generated by a CPU-based task:

1. Configure the peripheral and interrupt handler for the peripheral
2. CPU fills all three buffers with data to be used by peripheral
3. CPU writes address of buffer #1 into register 'PTR'
4. CPU writes value into TASK_STARTED register (starts the DMA)
5. (peripheral copies PTR register's value, aka buffer #1, to internal register)
6. (peripheral generates EVENTS_STARTED event)
7. INT HANDLER: EVENTS_STARTED: Set buffer #2 into register 'PTR' && exit
8. (DMA continues)
9. (peripheral finishes write to buffer #1 ... generates EVENTS_END event)
10. (peripheral copies PTR register's value, aka buffer #2, to internal register)
11. (peripheral generates EVENTS_STARTED event)
12. INT HANDLER: EVENTS_STARTED and EVENTS_END are both (potentially) set
    a. EVENTS_END:     **Signals CPU task to consumer buffer #1** && exit
     b. EVENTS_STARTED: **Set buffer #3 into register 'PTR'**
13. (DMA continues)
14. CPU task fills buffer #1 with next set of data
15. **CPU task marks the buffer as ready for use by DMA engine**
16. << loop steps 7..14 until all data is sent to DMA engine, only buffer numbers changes >>

In particular, at step #11.a it may be somewhat complex to track which buffer had
just completed DMA.  The first time through, the value in register 'PTR' is buffer #2.
Thus, the address of the buffer where DMA is finished is NOT stored in any of the
peripheral's registers ... it must be tracked externally.

Q: does this need to consider, at step #11.a, if need to be able to detect and properly
handle a situation where two buffers are completed, but only one event is signaled?
e.g., if two EVENTS_END events fire in rapid succession and merge?

### Example Peripherals

Long-term, this library is intended to make it easier to support using DMA
with (at least) the following peripherals:

* nRF52 I2S   (Inter-IC Sound Interface)
* nRF52 PDM   (Pulse Density Modulation)
* nRF52 PWM   (Pulse Width Modulation)
* nRF52 SAADC (Successive Approximation Analog-to-Digital Converter)
* nRF52 SPIM  (Serial Peripheral Interface -- Master)
* nRF52 TWIM  (Two Wire Interface -- I2C compatible -- Master)
* nRF52 UARTE (Universal Asynchronous Receiver/Transmitter)
* nRF52 USBD  (Universal Serial Bus Device)

### Properties of the example peripherals

* Supports EasyDMA
* Has a method to initiate a DMA transfer, using the current address pointer
  (e.g., writing to a TASKS_START register, DMA from address in PTR register)
* Has a method to stop future DMA transfers
  (e.g., writing to a TASKS_STOP register)
  NOTE: Current DMA transfer is usually completed before this takes effect.
* Has a STOPPED event when peripheral has stopped
  (e.g., EVENT_STOPPED)
* Has an event indicated buffer address register can be safely updated
  (e.g., EVENT_STARTED, EVENT_RXPTRUPD, EVENT_TXFRAMESTART, etc.)
  NOTE: e.g, the hardware has copied the buffer address internally
* Either automatically begins a next DMA transfer, or automatically stops
  * e.g., I2S, PDM always being a next DMA transfer unless TASKS_STOPPED occurred
    (??? auto-stops if the buffer is set to nullptr ???)
  * e.g., PWM, SPIM, TWIM, UARTE, USBD can uses a 'shortcuts' register to configure
    what next action should be when DMA completes (e.g., STOP or START sequences)
  * e.g., SAADC can have a timer automatically trigger a next sample
