#include "ezdmahelper.h"
#include "hw_rng.h"

// ~2912 bytes per CircularBufferHelper
#define MY_BUFFER_SIZE                  512
#define MY_BUFFER_COUNT                   4
#define MY_BUFFER_ALIGNMENT (sizeof(void*))

typedef EzDmaHelper<MY_BUFFER_SIZE, MY_BUFFER_COUNT, MY_BUFFER_ALIGNMENT> EZ_DMA_HELPER;
EZ_DMA_HELPER helper;

// volatile
// AND only incremented in corresponding callback task
// AND only decremented in loop() task
static volatile uint32_t g_BuffersAvailableToConsume = 0;
static volatile uint32_t g_BuffersAvailableToProduce = 0;

void onBufferProducedCallback(void * pvParameters)
{
  static uint8_t readAvailableCount = 0;
  readAvailableCount++;
  __sync_fetch_and_add(&g_BuffersAvailableToConsume, 1);
}
void onBufferConsumedCallback(void * pvParameters)
{
  static uint8_t writeAvailableCount = 0;
  writeAvailableCount++;
  __sync_fetch_and_add(&g_BuffersAvailableToProduce, 1);
}

void setup(){
    Serial.begin(115200);
    while (!Serial) {
      ; // wait for serial port to connect. Needed for native USB port only
    }
    // prints title with ending line break
    Serial.println("EasyDMA Helper -- test sketch");
    Serial.flush();
    delay(1000);

    SimpleHacks::HW_RNG::begin();

    helper.begin();
    helper.onDataBufferConsumed(onBufferConsumedCallback,  nullptr);
    helper.onDataBufferProduced(onBufferProducedCallback,  nullptr);
    g_BuffersAvailableToConsume = helper.GetBufferCountReadyForConsumer();
    g_BuffersAvailableToProduce = helper.GetBufferCountReadyForProducer();
    Serial.println("Setup complete");
    Serial.flush();
    delay(1000);
}

void OutputLineWithCharAtIndex(char toPlace, uint8_t index)
{
  // output: space, idx0, space, idx1, ... idxN, space, ERROR_INDEX, \n
    auto bufferSize = MY_BUFFER_COUNT*2 + 3;
    char tmp[bufferSize];
    for (auto i = 0; i < bufferSize-1; i++) {
      tmp[i] = ' ';
    }
    tmp[bufferSize-1] = '\0';
    if (index < MY_BUFFER_COUNT) {
      tmp[ index*2 ] = toPlace;
    } else {
      tmp[ MY_BUFFER_COUNT * 2 ] = '%';
    }
    Serial.println(tmp);    
}


void loop() {
    static uint32_t loopCount = 0;
    loopCount++;
    uint8_t randomness = SimpleHacks::HW_RNG::get_uint8();
  
    // guaranteed to only get larger from callbacks...
    uint8_t buffersToConsumeCount = 0;
    void* buffersToConsume[MY_BUFFER_COUNT] = {0};

    while (0 != g_BuffersAvailableToConsume && (randomness & 0x1)) {
        randomness >>= 1;
        // interlocked/atomic decrement
        __sync_fetch_and_sub(&g_BuffersAvailableToConsume, 1);
        
        void* buf = helper.GetNextBufferForConsumer();
        
        OutputLineWithCharAtIndex('C', helper.GetIndexFromPointer(buf));

        // assert it's non-null?
        if (nullptr == buf) {
          while(1) {
            Serial.println("_ nullptr from GetNextBufferForConsumer()?");
            Serial.flush();
            delay(5000);
          }
        }
        buffersToConsume[buffersToConsumeCount] = buf;
        buffersToConsumeCount ++;
    }

    for (uint8_t i = 0; i < buffersToConsumeCount; i++) {
        void* buf = buffersToConsume[i];
        OutputLineWithCharAtIndex('c', helper.GetIndexFromPointer(buf));
        delay(200);
        helper.SetBufferConsumed(buf);
    }
    randomness >>= 1;

    // guaranteed to only get larger from callbacks...
    uint8_t buffersToProduceCount = 0;
    void* buffersToProduce[MY_BUFFER_COUNT] = {0};

    while (0 != g_BuffersAvailableToProduce && (randomness & 0x1)) {
        randomness >>= 1;
        // interlocked/atomic decrement
        __sync_fetch_and_sub(&g_BuffersAvailableToProduce, 1);

        void* buf = helper.GetNextBufferForProducer();
        OutputLineWithCharAtIndex('P', helper.GetIndexFromPointer(buf));

        // assert it's non-null?
        if (nullptr == buf) {
          while(1) {
            Serial.println("_ nullptr from GetNextBufferForProducer()?");
            Serial.flush();
            delay(5000);
          }
        }
        buffersToProduce[buffersToProduceCount] = buf;
        buffersToProduceCount ++;
    }

    for (uint8_t i = 0; i < buffersToProduceCount; i++) {
        void* buf = buffersToProduce[i];
        OutputLineWithCharAtIndex('p', helper.GetIndexFromPointer(buf));
        helper.SetBufferProduced(buf);
    }
   

}
