/*
  Copyright (c) 2019 Henry Gabryjelski.  All right reserved.

  This file uses FreeRTOS features and headers.
  This file also requires C++11 (or later) due to the use of the following feature(s):

  [X] C++11 static members
      https://en.cppreference.com/w/cpp/language/static
*/

#ifndef _EZDMA_BUFFERS_H
#define _EZDMA_BUFFERS_H

#include <stddef.h> // 
#include <stdint.h>
#include <task.h> // freeRTOS tasks

#include <assert.h> // or, could use configASSERT(), except it's defined to be nothing...
#define LOCAL_ASSERT(X) assert(X)


namespace { // constexpr utility functions
    constexpr bool is_power_of_two(size_t v) {
        return v && ((v & (v - 1)) == 0);
    }
    constexpr bool is_power_of_two(uint32_t v) {
        return v && ((v & (v - 1)) == 0);
    }
    constexpr bool is_power_of_two(uint16_t v) {
        return v && ((v & (v - 1)) == 0);
    }
    constexpr bool is_power_of_two(uint8_t v) {
        return v && ((v & (v - 1)) == 0);
    }
    template<typename T> constexpr T const& max(T const& a, T const& b) {
        return a > b ? a : b;
    }
    template<typename T> constexpr T const& min(T const& a, T const& b) {
        return a > b ? a : b;
    }
};


#define DEFAULT_EZDMA_BUFFER_SIZE      512
#define DEFAULT_EZDMA_BUFFER_COUNT       2
#define DEFAULT_EZDMA_BUFFER_ALIGNMENT sizeof(uint8_t)
#define DEFAULT_TASK_PRIORITY             configTIMER_TASK_PRIORITY
static_assert(
    DEFAULT_EZDMA_BUFFER_COUNT != 0,
    "Default buffer count must be non-zero"
    );
static_assert(
    is_power_of_two(DEFAULT_EZDMA_BUFFER_ALIGNMENT),
    "Default alignment must be power of two, else __alignof() would be ill-formed"
    );
static_assert(
    0 == (DEFAULT_EZDMA_BUFFER_SIZE % DEFAULT_EZDMA_BUFFER_ALIGNMENT),
    "Default buffer size must be integral multiple of alignment, else gaps between buffers will exist"
    );
static_assert(
    DEFAULT_EZDMA_BUFFER_COUNT * DEFAULT_EZDMA_BUFFER_SIZE >= DEFAULT_EZDMA_BUFFER_SIZE,
    "Overflow default buffer size * default buffer count"
    );

typedef enum _INTERNAL_NOTIFICATION_TYPE { // bitmask
    InternalNotificationTypeUndefined     = 0x00000000,
    InternalNotificationTypeBufferConsumed    = 0x00000001, // queue == 0, 1<<0 == 1
    // Reserved                           = 0x00000002, // queue == 1, 1<<1 == 2
    InternalNotificationTypeBufferProduced = 0x00000004, // queue == 2, 1<<2 == 4
    // Reserved                           = 0x00000008, // queue == 3, 1<<3 == 8
} INTERNAL_NOTIFICATION_TYPE;


// A clean implementation would use a task for callbacks:
// * Create the task on first callback registration
// * Task handle is saved to private member variable
// * Task blocks on xTaskNotifyWait( pdFALSE, ULONG_MAX, &notificationValue, INFINITE )
// * Task calls the callback(s) indicated by the notification value
// * ISRs etc. call xTaskNotifyAndQuery() / xTaskNotifyAndQueryFromIsr()
//   to set bits notifying that buffers are now available.
// * Task Priority: configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY + 1 (aka 3 on Adafruit nRF52)

typedef void (*DataBufferProducedCallback)(void* context);
typedef void (*DataBufferConsumedCallback)(void* context);

template<uint32_t BUFF_SIZE, uint8_t BUFF_COUNT, uint32_t BUFF_ALIGNMENT>
void EZDMABufferNotificationTask( void * pvParameters );

static const char * queueNames[] = {
    "EZDMA_ReadyForProducing",
    "EZDMA_ProducingData",
    "EZDMA_ReadyForUse",
    "EZDMA_UsingData"
    };




// This is the base template that can be used by any combination of producers/consumers...
// whether they are DMA-enabled hardware, or CPU-based tasks.
// However, this is somewhat heavy overhead if at least one of those is not DMA-enabled hardware....
template <
    // the count of bytes per buffer
    uint32_t EZDMA_BUFFER_SIZE = DEFAULT_EZDMA_BUFFER_SIZE,
    // the number of such buffers to have available
    uint8_t  EZDMA_BUFFER_COUNT = DEFAULT_EZDMA_BUFFER_COUNT,
    // alignment requirement for the buffer (e.g., if DMA engine requires 4-byte alignment, this would be 4)
    size_t EZDMA_BUFFER_ALIGNMENT = DEFAULT_EZDMA_BUFFER_ALIGNMENT,
    // allow setting the priority for a task
    UBaseType_t EZDMA_BUFFER_TASK_PRIORITY = DEFAULT_TASK_PRIORITY
    >
class EzDmaHelper
    final // while it's OK to encapsulate this class, this class should not be directly derived from (see special members)
{
public: // static assertions
    // Note: these match those applied to the defaults
    static_assert(
        EZDMA_BUFFER_COUNT != 0,
        "Buffer count must be non-zero"
        );
    static_assert(
        is_power_of_two(EZDMA_BUFFER_ALIGNMENT),
        "Alignment must be power of two, else __alignof() would be ill-formed"
        );
    static_assert(
        0 == (EZDMA_BUFFER_SIZE % EZDMA_BUFFER_ALIGNMENT),
        "Buffer size must be integral multiple of alignment, else gaps between buffers will exist"
        );
    static_assert(
        EZDMA_BUFFER_COUNT * EZDMA_BUFFER_SIZE >= EZDMA_BUFFER_SIZE,
        "Overflow buffer size * buffer count"
        );
public: // constructor / destructor
    EzDmaHelper() {
        // APIs used in the constructor selected so that no failure paths exist
        // xTaskCreateStatic() cannot fail if neither puxStackBuffer nor pxTaskBuffer are nullptr
        m_xHandle = xTaskCreateStatic(
            EzDmaBufferNotificationTask,
            "EzDmaNo",
            c_TaskStackSize,
            this,
            EZDMA_BUFFER_TASK_PRIORITY,
            m_StaticTaskStack,
            &m_StaticTaskBuffer
            );
        // xQueueCreateStatic cannot fail initialization
        for (auto i = 0; i < 4; i++) {
            m_Queue[i] = xQueueCreateStatic(
                EZDMA_BUFFER_COUNT, // cannot queue more than we allocate
                sizeof(uint8_t),       // queue stores the buffer index only
                m_StaticQueueStorageArea[i],
                &(m_StaticQueueAllocation[i])
                );
            vQueueAddToRegistry(m_Queue[i], queueNames[i]);
        }
    };
    ~EzDmaHelper() {
        // APIs used in the destructor selected so that no failure paths exist
        // TODO: validate no buffers in queues 1 & 3 (in-process of reading/writing)
        for (auto i = 0; i < 4; i++) {
            if (nullptr != m_Queue[i]) {
                vQueueUnregisterQueue(m_Queue[i]);
                m_Queue[i] = nullptr;
            }
        }
        if (nullptr != m_xHandle) { // just in case ... since passing nullptr would delete the currently-executing task!
            vTaskDelete(m_xHandle);
            m_xHandle = nullptr;
        }
    };
public: // special members -- explicitly disallowing copy / assignment
    EzDmaHelper(
        const EzDmaHelper<EZDMA_BUFFER_SIZE,EZDMA_BUFFER_COUNT, EZDMA_BUFFER_ALIGNMENT, EZDMA_BUFFER_TASK_PRIORITY>&
        ) =delete;
    EzDmaHelper<EZDMA_BUFFER_SIZE,EZDMA_BUFFER_COUNT, EZDMA_BUFFER_ALIGNMENT, EZDMA_BUFFER_TASK_PRIORITY>& operator=(
        const EzDmaHelper<EZDMA_BUFFER_SIZE,EZDMA_BUFFER_COUNT, EZDMA_BUFFER_ALIGNMENT, EZDMA_BUFFER_TASK_PRIORITY>&
        ) =delete;
    EzDmaHelper(
        EzDmaHelper<EZDMA_BUFFER_SIZE,EZDMA_BUFFER_COUNT, EZDMA_BUFFER_TASK_PRIORITY, EZDMA_BUFFER_ALIGNMENT>&&
        ) =delete;
    EzDmaHelper<EZDMA_BUFFER_SIZE,EZDMA_BUFFER_COUNT, EZDMA_BUFFER_TASK_PRIORITY, EZDMA_BUFFER_ALIGNMENT>& operator=(
        EzDmaHelper<EZDMA_BUFFER_SIZE,EZDMA_BUFFER_COUNT, EZDMA_BUFFER_TASK_PRIORITY, EZDMA_BUFFER_ALIGNMENT>&&
        ) =delete;

public: // Public APIs -- begin() and end() -- Arduino API Style Guide
    void begin() {
        // clear the buffers
        memset(m_Buffer, 0, EZDMA_BUFFER_COUNT * EZDMA_BUFFER_SIZE);

        // sanity-check that the queues don't have buffers "in use"
        LOCAL_ASSERT( 0 == GetBufferCountCurrentlyBeingProduced() );
        LOCAL_ASSERT( 0 == GetBufferCountCurrentlyBeingConsumed() );

        // reset the queues
        for (auto i = 0; i < 4; i++) {
            BaseType_t result = xQueueReset(m_Queue[i]);
            LOCAL_ASSERT(pdTRUE == result);
        }

        // add all the buffers to the "ready to write" queue
        auto q = m_Queue[0];
        for (uint8_t i = 0; i < EZDMA_BUFFER_COUNT; i++) {
            BaseType_t result = xQueueSendToBack(q, &i, 0);
            LOCAL_ASSERT(pdTRUE == result);
        }
    }
    void end() {}
public: // Public APIs -- onDataBufferProduced() and onDataBufferConsumed() -- client callbacks
    // NOTE: callback may be called when there are zero, one, or more buffers with data available to consume:
    //       callback function must check for and process all available buffers.
    //       callback function must handle case where no buffer is available.
    bool onDataBufferProduced(DataBufferProducedCallback callback, void* context) {
        taskENTER_CRITICAL();
        m_DataBufferProducedCallback = callback;
        m_DataBufferProducedCallbackContext = context;
        taskEXIT_CRITICAL();
        return true;
    }
    // NOTE: callback may be called when there are zero, one, or more buffers available to write data to:
    //       callback function must check for and process all available buffers.
    //       callback function must handle case where no buffer is available.
    bool onDataBufferConsumed(DataBufferConsumedCallback callback, void* context) {
        taskENTER_CRITICAL();
        m_DataBufferConsumedCallback = callback;
        m_DataBufferConsumedCallbackContext = context;
        taskEXIT_CRITICAL();
        return true;
    }
private: // internal APIs (client callbacks)
    static void EzDmaBufferNotificationTask(void * pvParameters) {
        EzDmaHelper<EZDMA_BUFFER_SIZE, EZDMA_BUFFER_COUNT, EZDMA_BUFFER_ALIGNMENT, EZDMA_BUFFER_TASK_PRIORITY> * instance =
            static_cast<EzDmaHelper<EZDMA_BUFFER_SIZE, EZDMA_BUFFER_COUNT, EZDMA_BUFFER_ALIGNMENT, EZDMA_BUFFER_TASK_PRIORITY> *>(pvParameters);
        for (;;)
        {
            uint32_t notificationValue = 0;
            xTaskNotifyWait(
                0,                   // do not clear any notification bits on function entry
                UINT32_MAX,          // clear all notification bits as the wait ends
                &notificationValue,  // stores notification bits to be acted upon
                portMAX_DELAY        // wait forever...
                );
            if ((notificationValue & InternalNotificationTypeBufferProduced) != 0) {
                taskENTER_CRITICAL();
                DataBufferProducedCallback callback = instance->m_DataBufferProducedCallback;
                void* context = instance->m_DataBufferProducedCallbackContext;
                taskEXIT_CRITICAL();
                callback(context);
            }
            if ((notificationValue & InternalNotificationTypeBufferConsumed) != 0) {
                taskENTER_CRITICAL();
                DataBufferConsumedCallback callback = instance->m_DataBufferConsumedCallback;
                void* context = instance->m_DataBufferConsumedCallbackContext;
                taskEXIT_CRITICAL();
                callback(context);
            }
        }
    }
public: // Public APIs -- Buffer management functions
    // [ ] Returns next buffer pointer, or nullptr when no futher buffer is available.
    // [ ] Separate functions are provided to allow calling from an interrupt service routine.
    void* GetNextBufferForProducer(TickType_t xTicksToWait = 0) {
        void* actual = internal_moveBufferBetweenQueues(0, 1, xTicksToWait);
        if (xTicksToWait == portMAX_DELAY) {
            LOCAL_ASSERT(nullptr != actual);
        }
        return actual;
    }
    void* GetNextBufferForProducerFromISR() {
        return internal_moveBufferBetweenQueuesFromISR(0,1);
    }
    // [ ] Callers will preferably use `SetBufferProduced(buffer)`,
    //     when the caller has the pointer to the completed buffer at-hand,
    //     as it allows validation that the buffers are continuing
    //     to be processed in-order.  However, the function is also
    //     provided with no argument, for when the buffer pointer
    //     is not easily kept, but the buffers must still be written
    //     to in order.
    // [ ] *** Caller must ensure buffers are completed in FIFO order.
    //     This is easily supported with EasyDMA style peripherals,
    //     such as are common on the nRF5x series.
    // [ ] Separate functions are provided to allow calling from
    //     an interrupt service routine.
    void SetBufferProduced() {
        void* actual = internal_moveBufferBetweenQueues(1, 2);
        LOCAL_ASSERT(nullptr != actual);
    }
    void SetBufferProduced(void* buffer) {
        LOCAL_ASSERT(nullptr != buffer);
        void * actual = internal_moveBufferBetweenQueues(1, 2);
        LOCAL_ASSERT(actual == buffer);
    }
    void SetBufferProducedFromISR() {
        void * actual = internal_moveBufferBetweenQueuesFromISR(1, 2);
        LOCAL_ASSERT(nullptr != actual);
    }
    void SetBufferProducedFromISR(void* buffer) {
        LOCAL_ASSERT(nullptr != buffer);
        void* actual = internal_moveBufferBetweenQueuesFromISR(1, 2);
        LOCAL_ASSERT(actual == buffer);
    }
    // [ ] Returns next buffer pointer, or nullptr when no futher buffer
    //     is available.
    // [ ] Separate functions are provided to allow calling from
    //     an interrupt service routine.
    void* GetNextBufferForConsumer(TickType_t xTicksToWait = 0) {
        void* actual = internal_moveBufferBetweenQueues(2, 3, xTicksToWait);
        if (xTicksToWait == portMAX_DELAY) {
            LOCAL_ASSERT(nullptr != actual);
        }
        return actual;
    }
    void* GetNextBufferForConsumerFromISR() {
        return internal_moveBufferBetweenQueuesFromISR(2, 3);
    }
    // [ ] Callers will preferably use `SetReadCompleted(buffer)`,
    //     when the caller has the pointer to the completed buffer at-hand,
    //     as it allows validation that the buffers are continuing
    //     to be processed in-order.  However, the function is also
    //     provided with no argument, for when the buffer pointer
    //     is not easily kept, but the buffers must still be written
    //     to in order.
    // [ ] *** Caller must ensure buffers are completed in FIFO order.
    //     This is easily supported with EasyDMA style peripherals,
    //     such as are common on the nRF5x series.
    // [ ] Separate functions are provided to allow calling from
    //     an interrupt service routine.
    void SetBufferConsumed() {
        void* actual = internal_moveBufferBetweenQueues(3, 0);
        LOCAL_ASSERT(nullptr != actual);
    }
    void SetBufferConsumed(void* buffer) {
        LOCAL_ASSERT(nullptr != buffer);
        void* actual = internal_moveBufferBetweenQueues(3, 0);
        LOCAL_ASSERT(actual == buffer);
    }
    void SetBufferConsumedFromIsr() {
        void * actual = internal_moveBufferBetweenQueuesFromISR(3, 0);
        LOCAL_ASSERT(nullptr != actual);
    }
    void SetBufferConsumedFromIsr(void* buffer) {
        LOCAL_ASSERT(nullptr != buffer);
        void* actual = internal_moveBufferBetweenQueuesFromISR(3, 0);
        LOCAL_ASSERT(actual == buffer);
    }
public: // Public APIs -- Get buffer index from pointer
    uint8_t GetIndexFromPointer(void* buffer) {
        uintptr_t toTest = (uintptr_t)buffer;
        uintptr_t base = (uintptr_t)m_Buffer;
        if (toTest < base) return EZDMA_BUFFER_COUNT; // this is an error condition
        uintptr_t idx = (toTest - base) / EZDMA_BUFFER_SIZE;
        if (idx >= EZDMA_BUFFER_COUNT) return EZDMA_BUFFER_COUNT;
        if (base + (EZDMA_BUFFER_SIZE * idx) != toTest) return EZDMA_BUFFER_COUNT;
        return idx;
    }
    inline uint32_t    GetBufferSize()         { return EZDMA_BUFFER_SIZE;          }
    inline size_t      GetBufferAlignment()    { return EZDMA_BUFFER_ALIGNMENT;     }
    inline uint8_t     GetBufferCount()        { return EZDMA_BUFFER_COUNT;         }
    inline UBaseType_t GetBufferTaskPriority() { return EZDMA_BUFFER_TASK_PRIORITY; }
public: // Public APIs -- query queue counts
    // [ ] Returns count of buffers that are ready for writing
    uint8_t GetBufferCountReadyForProducer() {
        UBaseType_t result = uxQueueMessagesWaiting(m_Queue[0]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    uint8_t GetBufferCountReadyForProducerFromISR() {
        UBaseType_t result = uxQueueMessagesWaitingFromISR(m_Queue[0]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    // [ ] Returns count of buffers that are potentially in use for writing
    uint8_t GetBufferCountCurrentlyBeingProduced() {
        UBaseType_t result = uxQueueMessagesWaiting(m_Queue[1]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    uint8_t GetBufferCountCurrentlyBeingProducedFromISR() {
        UBaseType_t result = uxQueueMessagesWaitingFromISR(m_Queue[1]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    // [ ] Returns count of buffers that are ready for reading
    uint8_t GetBufferCountReadyForConsumer() {
        UBaseType_t result = uxQueueMessagesWaiting(m_Queue[2]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    uint8_t GetBufferCountReadyForConsumerFromISR() {
        UBaseType_t result = uxQueueMessagesWaitingFromISR(m_Queue[2]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    // [ ] Returns count of buffers that are potentially in use for reading
    uint8_t GetBufferCountCurrentlyBeingConsumed() {
        UBaseType_t result = uxQueueMessagesWaiting(m_Queue[3]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
    uint8_t GetBufferCountCurrentlyBeingConsumedFromISR() {
        UBaseType_t result = uxQueueMessagesWaitingFromISR(m_Queue[3]);
        LOCAL_ASSERT(result <= EZDMA_BUFFER_COUNT);
        return (uint8_t)result;
    }
private: // internal buffer management functions
    void* internal_moveBufferBetweenQueues(uint8_t fromQueue, uint8_t toQueue, TickType_t xTicksToWait = 0) {
        // NOTE: this function uses indices to the actual buffer pointer, not the actual buffer pointer itself
        uint8_t index = 0xFF;
        
        // First, get the next buffer index from the "fromQueue"
        BaseType_t result = xQueueReceive(m_Queue[fromQueue], &index, xTicksToWait);
        if (pdTRUE != result) {
            return nullptr;
        }
        LOCAL_ASSERT(index < EZDMA_BUFFER_COUNT);
        
        // Then, place the received buffer index into the "toQueue"
        result = xQueueSendToBack(m_Queue[toQueue], &index, 0);
        LOCAL_ASSERT(pdTRUE == result);

        // There are event notifications when placing a buffer onto
        // either the "ReadyForReading" or "ReadyForWriting" queues
        if ((toQueue == 0) || (toQueue == 2)) {
            uint32_t value = 1u << toQueue;
            xTaskNotify(m_xHandle, value, eSetBits);
        }

        // Finally, return the actual buffer pointer (not the index) when successful
        return m_Buffer[index];
    }
    void* internal_moveBufferBetweenQueuesFromISR(uint8_t fromQueue, uint8_t toQueue) {

        BaseType_t xHigherPriorityTaskWoken = pdFALSE;

        // NOTE: this function uses indices to the actual buffer pointer, not the actual buffer pointer itself
        uint8_t index = 0xFF;

        // First, get the next buffer index from the "fromQueue"
        BaseType_t result = xQueueReceiveFromISR(m_Queue[fromQueue], &index, &xHigherPriorityTaskWoken);
        void* retVal = nullptr;
        if (pdTRUE == result) {
            LOCAL_ASSERT(index < EZDMA_BUFFER_COUNT);
            
            // Then, place the received buffer index into the "toQueue"
            result = xQueueSendToBackFromISR(m_Queue[toQueue], &index, &xHigherPriorityTaskWoken);
            LOCAL_ASSERT(pdTRUE == result);

            // the returned value from the function is the actual buffer pointer
            retVal = m_Buffer[index];
        }
        
        // There are event notifications when placing a buffer onto
        // either the "ReadyForReading" or "ReadyForWriting" queues
        if ((toQueue == 0) || (toQueue == 2)) {
            uint32_t value = 1u << toQueue;
            xTaskNotifyFromISR(m_xHandle, value, eSetBits, &xHigherPriorityTaskWoken);
        }
        
        //// TODO: Determine if call to portYIELD_FROM_ISR(xHigherPriorityTaskWoken) should occur?
        //// If a higher priority task was woken by either the modification of either queue,
        //// ensure the higher priority task is woken immediately upon return from ISR
        // portYIELD_FROM_ISR(xHigherPriorityTaskWoken);

        // Finally, return the actual buffer pointer (not the index) when successful
        return retVal;
    }

private:
    // Client callbacks
    DataBufferProducedCallback  m_DataBufferProducedCallback          = nullptr;
    void *                      m_DataBufferProducedCallbackContext   = nullptr;
    DataBufferConsumedCallback  m_DataBufferConsumedCallback          = nullptr;
    void *                      m_DataBufferConsumedCallbackContext   = nullptr;

    // A single task for callbacks
    TaskHandle_t m_xHandle;
    // four queues -- ReadyForWrite, Writing, ReadyForRead, Reading
    QueueHandle_t m_Queue[4];

private: // static allocations (backing store for queues, DMA-aligned buffer, tasks, ...)
    // m_Queue backing memory:
    StaticQueue_t m_StaticQueueAllocation[4];
    uint8_t m_StaticQueueStorageArea[4][EZDMA_BUFFER_COUNT * sizeof(uint8_t)];

    // m_xHandle backing memory:
    static const uint32_t c_TaskStackSize = configMINIMAL_STACK_SIZE;
    StaticTask_t m_StaticTaskBuffer;
    StackType_t m_StaticTaskStack[c_TaskStackSize];

    // m_Buffer backing memory:
    // This is statically-allocated, and uses alignas() with the template argument to ensure
    // the buffer is aligned as per the DMA-engine requirements.  Thus, this effectively supports DMA
    // engines that require 
    // this allows callers to statically allocate this instance, thus avoid heap fragmentation
    alignas(EZDMA_BUFFER_ALIGNMENT)
    uint8_t m_Buffer[EZDMA_BUFFER_COUNT][EZDMA_BUFFER_SIZE];
};


// Example template for common case where the peripheral generates
// or otherwise obtains the data via DMA, with code to simplify
// reading the resulting data into arbitrary size buffers.
template<
    uint32_t BUFFER_SIZE=DEFAULT_EZDMA_BUFFER_SIZE,
    uint8_t  BUFFER_COUNT=DEFAULT_EZDMA_BUFFER_COUNT,
    size_t   BUFFER_ALIGNMENT=DEFAULT_EZDMA_BUFFER_ALIGNMENT,
    UBaseType_t BUFFER_TASK_PRIORITY=DEFAULT_TASK_PRIORITY
    >
class EzDma_Foo
{
public: // static assertions
    static_assert(
        BUFFER_COUNT != 0,
        "Buffer count must be non-zero"
        );
    static_assert(
        is_power_of_two(BUFFER_ALIGNMENT),
        "Alignment must be power of two, else __alignof() would be ill-formed"
        );
    static_assert(
        0 == (BUFFER_SIZE % BUFFER_ALIGNMENT),
        "Buffer size must be integral multiple of alignment, else gaps between buffers will exist"
        );
    static_assert(
        BUFFER_COUNT * BUFFER_SIZE >= BUFFER_SIZE,
        "Overflow buffer size * buffer count"
        );
public: // special members
    // standard constructor and destructor use default implementation
    EzDma_Foo() =default;
    ~EzDma_Foo() =default;
    // Disallow copy operator, assignment operator, move constructor, and move assignment operator
    EzDma_Foo(const EzDma_Foo<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY>&) =delete;
    EzDma_Foo<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY>& operator=(
        const EzDma_Foo<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY>&
        ) =delete;
    EzDma_Foo(EzDma_Foo<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY>&&) =delete;
    EzDma_Foo<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY>& operator=(
        EzDma_Foo<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY>&&
        ) =delete;

public: // Arduino API Style Guid based begin/end APIs
    void begin() {
        m_BufferHelper.begin();
    }
    void end() {}
    // expose smaller subset of APIs, targeting only this scenario
    void* GetNextBufferForWrite(BaseType_t xTicksToWait = 0)        { return m_BufferHelper.GetNextBufferForWrite(xTicksToWait); }
    void* GetNextBufferForWriteFromIsr()                            { return m_BufferHelper.GetNextBufferForWriteFromIsr();      }
    void  SetWriteCompleted()                                       { return m_BufferHelper.SetWriteCompleted();                 }
    void  SetWriteCompletedFromIsr()                                { return m_BufferHelper.SetWriteCompletedFromIsr();          }
    bool  onBytesAvailable(DataBufferProducedCallback callback, void* context) {
        return m_BufferHelper.onDataBufferProduced(callback, context);
    }
    size_t BytesAvailable() {
        size_t result = m_BufferHelper.GetBufferCountReadyForConsumer();
        result *= BUFFER_SIZE;
        if (m_CurrentDataBuffer) {
            LOCAL_ASSERT(m_BufferHelper.GetBufferCountCurrentlyBeingConsumed() != 0); // Sanity check state
            result += BUFFER_SIZE - m_CurrentDataBufferOffset;
        }
        return result;
    }
    size_t Read(void* buffer, size_t bytesToRead) {
        // NOTES:
        // [ ] This call is NON-BLOCKING.
        // [ ] May return ZERO if no buffer was available
        // [ ] May return LESS THAN bytesToRead ARGUMENT, if insufficient buffers available.
        // 
        // If caller requires more than one buffer of data from the call, they may
        // 1. loop calling BytesAvailableToRead() until required bytes are available
        // 2. call set_onBufferWritten() to register a callback for when a buffer is available to read
        //
        // It's not clear how to architect this to be blocking.
        // Allowing a wait time (even if only allowing infinite wait) is problematic because,
        // if the caller requests more data than will be generated, the API would hang waiting forever.
        //
        // Therefore, only a non-blocking API is provided.
        if (bytesToRead == 0) {
            return 0; // buffer is technically allowed to be nullptr if reading zero bytes...
        }
        LOCAL_ASSERT(buffer != nullptr);
        size_t bytesAlreadyRead = 0;

        do {
            // get buffer if one is not already available...
            if (nullptr == m_CurrentDataBuffer) {
                // this will be a blocking operation if caller asks to wait for the requested data
                m_CurrentDataBuffer = m_BufferHelper.GetNextBufferForConsumer();
                m_CurrentDataBufferOffset = 0;
            }
            // if still trying to read, but no more buffers, exit
            if (nullptr == m_CurrentDataBuffer) {
                break; // out of do{}while() loop
            }

            // copy bytes from the current buffer
            size_t bytesToCopy = min(bytesToRead, BUFFER_SIZE - m_CurrentDataBufferOffset);
            void*  copyFrom = m_CurrentDataBuffer + m_CurrentDataBufferOffset;
            memcpy(buffer, copyFrom, bytesToCopy);
            m_CurrentDataBufferOffset += bytesToCopy;
            bytesAlreadyRead += bytesToCopy;
            bytesToRead -= bytesToCopy;

            // if that exhausted the current buffer, return it to the queues
            if (m_CurrentDataBufferOffset >= BUFFER_SIZE) {
                LOCAL_ASSERT(m_CurrentDataBufferOffset == BUFFER_SIZE);
                void* buff = m_CurrentDataBuffer;
                m_CurrentDataBuffer = nullptr;
                m_CurrentDataBufferOffset = 0;
                m_BufferHelper.SetBufferConsumed(buff); // NOTE: may trigger callbacks
            }
        } while (bytesToRead);

        return bytesAlreadyRead;        
    }

private:
    // the buffer helper underlying this use....
    EzDmaHelper<BUFFER_SIZE,BUFFER_COUNT,BUFFER_ALIGNMENT,BUFFER_TASK_PRIORITY> m_BufferHelper;

    // at most, this class keeps one buffer pointer cached for reading...
    void*  m_CurrentDataBuffer;
    size_t m_CurrentDataBufferOffset; // how much data has already been read...
};

// TODO:
// [ ] Consider if should assert, if Read() related functions are called from different tasks
//     PROS: might avoid common race conditions
//     CONS: might prevent valid use cases
//     Affected Functions:
//         BytesAvailableToRead()
//         Read()
//
//     void DEBUG_SetOrVerifySameTaskHandle(TaskHandle& toCheck) {
//     #if INCLUDE_xTaskGetCurrentTaskHandle
//         TaskHandle_t self = xTaskGetCurrentHandle();
//         if (nullptr == toCheck) {
//             toCheck = self;
//         } else {
//             LOCAL_ASSERT(self == toCheck);
//         }
//     #endif // INCLUDE_xTaskGetCurrentTaskHandle
//     }
//     TaskHandle_t m_read_functions_task_handle;
//     DEBUG_SetOrVerifySameTaskHandle(&m_function_task_handle);
//     begin() would also clear this handle to nullptr
//


#endif // _EZDMA_BUFFERS_H
